#include <cuda.h>
#include <cuda_runtime.h>
#include <opencv.hpp>
#include <device_functions.h>
#include <stdio.h>
#include <iostream>
#include <sys/time.h>

static void HandleError( cudaError_t err,
                         const char *file,
                         int line ) {
    if (err != cudaSuccess) {
        printf( "%s in %s at line %d\n", cudaGetErrorString( err ),
                file, line );
        exit( EXIT_FAILURE );
    }
}
static bool firstRun = true;
static uchar *popcount16LUT = new uchar[65536];
static const int dispnum = 64;

#define HANDLE_ERROR( err ) (HandleError( err, __FILE__, __LINE__ ))


#define HANDLE_NULL( a ) {if (a == NULL) { \
                            printf( "Host memory failed in %s at line %d\n", \
                                    __FILE__, __LINE__ ); \
                            exit( EXIT_FAILURE );}}

#define  WIDTH   1024
#define  HEIGHT  400


static unsigned char hamDist_16(ushort x, ushort y)
{
    unsigned char dist = 0;
    unsigned short val = (x^y);

    while(val)
    {
        ++dist;
        val &= val - 1;
    }

    return dist;
}


__global__ void TransposeDSI(unsigned short *dsiSrc, unsigned short *dsiDes, int width, int height, int n_disp)
{
    int i = threadIdx.x;
    int j = blockIdx.x;
    int index1, index2;
    int d;

    index1=j*width*n_disp+i*n_disp;
    index2=i*height*n_disp+j*n_disp;
    dsiDes += index2;
    dsiSrc += index1;
    for(d = 0; d < n_disp; d++) {
        dsiDes[d] = dsiSrc[d];
        dsiSrc[d] = 0;
    }
    //memcpy(dsiDes+index2, dsiSrc+index1, sizeof(unsigned short)*n_disp);
}





__global__ void getCensus(unsigned char *left_data, unsigned char *right_data,
                          unsigned char *popcount16LUT, unsigned short *dsi,
                          int height, int width)
{
    __shared__ unsigned long long int leftCensus[1024];
    __shared__ unsigned long long int rightCensus[1024];
    unsigned long long int left_val = 0;
    unsigned long long int right_val = 0;
    unsigned long long int val_l;
    unsigned long long int val_r;



    int tid = threadIdx.x;
    int bid = blockIdx.x;


    if(tid >= 4 && bid >= 3) {

        unsigned char left_center = left_data[bid*width + tid];
        unsigned char right_center = right_data[bid*width + tid];

        for (int m = -3; m <= 3; m++)
        {
            for (int n =  -4; n <= 4; n++)
            {
                left_val = (left_val << 1);
                left_val += (left_data[(bid+m) * width + tid + n] < left_center);

                right_val = (right_val << 1);
                right_val += (right_data[(bid+m) * width + tid + n] < right_center);
            }
        }
        leftCensus[tid] = left_val;
        rightCensus[tid] = right_val;

        __syncthreads();
        for (int d = 0; d < 64; d++)
        {
            if( d <= tid - 4)
            {
                val_l = leftCensus[tid];
                val_r = rightCensus[tid-d];
            // total 44 bytes
                dsi[(bid*width+tid)*64 + d] = popcount16LUT[ (unsigned short)(val_l ^ val_r)]
                    + popcount16LUT[(unsigned short)((val_l>>16) ^ (val_r>>16))]
                    + popcount16LUT[(unsigned short)((val_l>>32) ^ (val_r>>32))];
            }
            else
            {
                dsi[(bid*width+tid)*64 + d] = 0xff;
            }
        }
    }
    else {
        leftCensus[tid] = 0;
        rightCensus[tid] = 0;
    }

}



uchar *left_data, *right_data;
float *disp_img;
unsigned short *out_dist;
unsigned short *dist_agg;


extern "C"
void cudaInit()
{
    HANDLE_ERROR(cudaMalloc( (void**)&left_data, WIDTH * HEIGHT * sizeof(uchar)));
    HANDLE_ERROR(cudaMalloc( (void**)&right_data, HEIGHT * WIDTH * sizeof(uchar)));
    HANDLE_ERROR(cudaMalloc( (void**)&out_dist, sizeof(unsigned short)*WIDTH*HEIGHT*dispnum));
    HANDLE_ERROR(cudaMalloc( (void**)&dist_agg, sizeof(unsigned short)*WIDTH*HEIGHT*dispnum));
    HANDLE_ERROR(cudaMalloc( (void**)&disp_img, WIDTH * HEIGHT *sizeof(float)));
    if(firstRun)
    {
        HANDLE_ERROR(cudaMalloc( (void**)&popcount16LUT, 65536));
        static uchar *tmp_popcount16LUT = new uchar[65536];
        // build the popcount16LUT
        for (int i = 0; i < 65536; i++) {
            tmp_popcount16LUT[i] = hamDist_16(i,0);
        }

        HANDLE_ERROR(cudaMemcpy(popcount16LUT, tmp_popcount16LUT, 65536, cudaMemcpyHostToDevice));
        firstRun = false;
    }
}


extern "C"
int cudaCensus(cv::Mat& left_img, cv::Mat& right_img, unsigned short *dsi)
{

    cudaDeviceProp prop;
    int dev;
    cudaGetDevice(&dev);
    cudaGetDeviceProperties(&prop,dev);

    printf("device name: %s\n", prop.name);
    printf("computer ability: %d.%d\n", prop.major, prop.minor);
    printf("MaxGridSize: %d,%d,%d\n", prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2]);
    printf("MaxThreadPerBlock: %d\n", prop.maxThreadsPerBlock);
    printf("MaxThreadDim: %d,%d,%d\n", prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2]);
    printf("multiProcessorCount: %d\n", prop.multiProcessorCount);
    printf("resPerBlock: %d(K)\n", prop.regsPerBlock);
    printf("sharedMemoryPerBlock(K): %d\n", prop.sharedMemPerBlock/1024);
    printf("totalGlobalMemory: %d(M)\n", prop.totalGlobalMem/(1024*1024));
    printf("warpSize: %d\n", prop.warpSize);
    printf("ConstanMemory: %d(K)\n", prop.totalConstMem/1024);


    int height = left_img.rows;
    int width  = left_img.cols;
    int data_size = height * width * sizeof(uchar);

    HANDLE_ERROR(cudaMemcpy(left_data, left_img.data, data_size, cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMemcpy(right_data, right_img.data, data_size, cudaMemcpyHostToDevice));

    cudaEvent_t getCensus_start, getCensus_stop;
    cudaEventCreate(&getCensus_start);
    cudaEventCreate(&getCensus_stop);
    cudaEventRecord(getCensus_start, NULL);

    getCensus<<<height, width>>>(left_data, right_data, popcount16LUT, out_dist, height, width);
    cudaDeviceSynchronize();

    cudaEventRecord(getCensus_stop, NULL);
    cudaEventSynchronize(getCensus_stop);
    float getCensus_time = 0.0f;
    cudaEventElapsedTime(&getCensus_time, getCensus_start, getCensus_stop);
    cudaEventDestroy(getCensus_start);
    cudaEventDestroy(getCensus_stop);
    std::cout << " getCensus  time=" << getCensus_time<<std::endl;


    /* TransposeDSI */
    cudaEvent_t TransposeDSI_1_start, TransposeDSI_1_stop;
    cudaEventCreate(&TransposeDSI_1_start);
    cudaEventCreate(&TransposeDSI_1_stop);
    cudaEventRecord(TransposeDSI_1_start, NULL);

    TransposeDSI<<<height, width>>>(out_dist, dist_agg, width, height, dispnum);
    cudaDeviceSynchronize();

    cudaEventRecord(TransposeDSI_1_stop, NULL);
    cudaEventSynchronize(TransposeDSI_1_stop);
    float TransposeDSI_1_time = 0.0f;
    cudaEventElapsedTime(&TransposeDSI_1_time, TransposeDSI_1_start, TransposeDSI_1_stop);
    cudaEventDestroy(TransposeDSI_1_start);
    cudaEventDestroy(TransposeDSI_1_stop);
    std::cout << " TransposeDSI time=" << TransposeDSI_1_time<<std::endl;

    HANDLE_ERROR(cudaMemcpy(dsi, dist_agg, width * height *sizeof(unsigned short), cudaMemcpyDeviceToHost));


    return 0;

}



